#! /usr/bin/perl -wT

package Twici ;

use strict ;

use Exporter ;
our @ISA = qw(Exporter) ;
our @EXPORT = qw(handleView) ;

### CONSTANTS

# todo: eliminate these where possible

use constant WIKI_ROOT => "/srv/www/urchin.earth.li/users/twic/wiki/" ;
use constant WIKI_EXT => ".twici" ;
use constant WIKI_SRV => "http://urchin.earth.li/" ;
use constant WIKI_BASE => WIKI_SRV . "cgi-bin/twic/wiki/" ;
use constant WIKI_VIEW_BASE => WIKI_BASE . "view.pl?page=" ;
use constant WIKI_EDIT_BASE => WIKI_BASE . "edit.pl?page=" ;
use constant WIKI_SOURCE_BASE => WIKI_BASE . "source.pl?page=" ;
use constant WIKI_BACKLINKS_BASE => WIKI_BASE . "backlinks.pl?page=" ;
use constant WIKI_CHANGES_ACTION => WIKI_BASE . "changes.pl" ;
use constant WIKI_NAME_RE => "([A-Z][0-9a-z]*[A-Z][0-9a-zA-Z]*)" ;
use constant BARE_WIKI_NAME_RE => "^&WIKI_NAME_RE\$" ;
use constant EMBEDDED_WIKI_NAME_RE => "^([^0-9a-zA-z]*)&WIKI_NAME_RE([^0-9a-zA-z]*)\$" ;
use constant WIKI_PAGE_FILENAME_RE => "^&WIKI_NAME_RE&WIKI_EXT\$" ;
use constant URL_RE => "([a-zA-Z][0-9a-zA-Z+\-\.]*:.+)" ;
use constant EMBEDDED_URL_RE => "^([^0-9a-zA-z]*)&URL_RE([^0-9a-zA-z]*)\$" ;
use constant FRONT_PAGE => "FrontPage" ;
use constant CHANGES_PAGE => "RecentChanges" ;
use constant MOTD_PAGE => "MessageOfTheDay" ;
use constant BADURL_PAGE => "BadURL" ;
use constant KEY_PAGES => (FRONT_PAGE, CHANGES_PAGE, MOTD_PAGE) ;
use constant SECS_PER_DAY => 24 * 60 * 60 ;
use constant DOCTYPE => "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">" ;

### VIEW

sub handleView
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = &parse( $queryString ) ;
	my $page = &cleanPage( $query{"page"} ) ;
	my $pageFile = &pageFile( $page ) ;
	open( PAGE, "< " . $pageFile ) || die( "could not open page file" ) ;
	print "Content-type: text/html\n" ;
	print "\n" ;
	print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>$page</title>
</head>
<body>
<h1>$page</h1>
<hr>
EOF
	$/ = "" ; # paragraph mode; note that we don't ever leave it again
	while (my $block = <PAGE>)
	{
		&format( $block ) ;
	}
	close( PAGE ) ;
	my $editUrl = &editUrl( $page ) ;
	my $backlinksUrl = &backlinksUrl( $page ) ;
	my $sourceUrl = &sourceUrl( $page ) ;
	my $validateUrl = &validateUrl( $page ) ;
	print <<EOF ;
<hr>
<p>
| <a href="$editUrl">Edit</a>
| <a href="$backlinksUrl">Backlinks</a>
| <a href="$sourceUrl">Source</a>
| <a href="$validateUrl">Validate</a> 
|
EOF
	foreach my $keyPage (&KEY_PAGES)
	{
		my $keyUrl = &viewUrl( $keyPage ) ;
		print "| <a href=\"$keyUrl\">$keyPage</a>\n" ;
	}
	my $mtimestr = gmtime( pageMtime( $page ) ) ;
	print <<EOF ;
|
| Last Modified $mtimestr GMT
|
</p>
</body>
</html>
EOF
}

# below be scary formatting related things
# step one in refactoring it: extract to method
# step two will be 'hit with cattle-prods until docile'

sub format
{
	my $block = shift @_ ;
	if (substr( $block, 0, 1 ) eq "-")
	{
		list( $block ) ;
	}
	else
	{
		paragraph( $block ) ;
	}
}

sub list
{
	my $list = shift @_ ;
	my @lines = split( "\n", $list ) ;
	print "<ul>\n" ;
	foreach my $line (@lines)
	{
		if (substr( $line, 0, 1 ) eq "-")
		{
			print "<li>" ;
			$line =~ /-(?:\s*)?(.*)/ ; # HNUUUUUHHHH!!
			$line = $1 ;
		}
		text( $line ) ;
		print "\n" ;
	}
	print "</ul>\n" ;
}

sub paragraph
{
	my $para = shift @_ ;
	print "<p>" ;
	text( $para ) ;
	print "</p>\n" ;
}

sub text
{
	my $text = shift @_ ;
	$text =~ s/</&lt;/g ;
	my @words = split( " ", $text ) ; # actually splits on whitespace
	foreach my $word (@words)
	{
		if (substr( $word, 0, 1 ) eq "_")
		{
			print "<em>" ;
			$word = substr( $word, 1 ) ;
		}
		elsif (substr( $word, 0, 1 ) eq "*")
		{
			print "<strong>" ;
			$word = substr( $word, 1 ) ;
		}
		my $trailingTag ;
                if (substr( $word, -1, 1 ) eq "_")
                {
                        $trailingTag = "</em>" ;
                        $word = substr( $word, 0, -1 ) ;
                }
                elsif (substr( $word, -1, 1 ) eq "*")
                {
                        $trailingTag = "</strong>" ;
                        $word = substr( $word, 0, -1 ) ;
                }
		my @parts ;
		if (scalar (@parts = extractWikiName( $word )))
		{
			my $pre = $parts[0] ;
			my $name = $parts[1] ;
			my $post = $parts[2] ;
			if (&exists( $name ))
			{
				my $href = WIKI_VIEW_BASE . $name ;
				print "$pre<a href=\"$href\">$name</a>$post " ;
			}
			else
			{
				my $href = WIKI_EDIT_BASE . $name ;
				print "$pre$name<a href=\"$href\">?</a>$post " ;
			}
		}
		elsif (scalar (@parts = extractUrl( $word )))
		{
                        my $pre = $parts[0] ;
                        my $url = $parts[1] ;
                        my $post = $parts[2] ;
			my $genUrl = generaliseUrl( $url ) ;
			print "$pre<a href=\"$genUrl\">$url</a>$post " ;
		}
		else
		{
			print "$word " ;
		}
		if (defined( $trailingTag ))
		{
			print $trailingTag ;
		}
	}
}

# todo: handle errors more robustly; maybe emit a link to a 'bad url' page
sub generaliseUrl
{
	my $url = shift @_ ;
	$url =~ /([A-Za-z][A-Za-z0-9\+\-\.]*):(.*)/ || return badUrl( $url, "invalid scheme") ;
	my $scheme = lc( $1 ) ;
	my $path = $2 ;
	if ($scheme eq "isfdb")
	{
		$path =~ /([A-Za-z]+)\/(.+)/ || return badUrl( $url, "invalid ISFDB URL" ) ;
		my $class = lc( $1 ) ;
		my $entity = $2 ;
		if ($class eq "author")
		{
			return "http://www.sfsite.com/isfdb-bin/exact_author.cgi?" . $entity ;
		}
		elsif ($class eq "work")
		{
			return "http://www.sfsite.com/isfdb-bin/pwork.cgi?" . $entity ;
		}
		else
		{
			return badUrl( $url, "invalid ISFDB class" ) ;
		}
	}
	elsif ($scheme eq "lj")
	{
		# should validate username
		return "http://www.livejournal.com/users/" . $path ;
	}
	elsif ($scheme eq "groggs")
	{
		# should validate ID
		return "gopher://gopher.publication.org.uk:70/1/rgtp/index%3Fgroggs/" . $path ;
	}
	elsif ($scheme eq "google")
	{
		return "http://www.google.co.uk/search?q=" . $path ;
	}
	else
	{
		return $url ;
	}
}

sub badUrl
{
	my $url = shift @_ ;
	my $msg = shift @_ ;
	if ($msg eq undef)
	{
		$msg = "" ;
	}
	warn "invalid URL ($msg): $url" ;
	return viewUrl( BADURL_PAGE ) ;
}

### EDIT

sub handleEdit
{

	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = parse( $queryString ) ;
	my $page = &cleanPage( $query{"page"} ) ;
	my $requestMethod = $ENV{"REQUEST_METHOD"} ;
	if ($requestMethod eq "GET")
	{
		&editPage( $page ) ;
	}
	else
	{
		$requestMethod eq "POST" || die( "bad request method: $requestMethod" ) ;
		&updatePage( $page ) ;
	}
}

sub editPage
{
	my $page = shift @_ ;
	my $pageFile = &pageFile( $page ) ;
	print "Content-type: text/html\n" ;
	print "\n" ;
	my $action = editUrl( $page ) ;
	print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>Edit $page</title>
</head>
<body>
<h1>$page</h1>
<hr>
<form action="$action" method="POST">
<textarea name="text" rows="20" cols="80">
EOF
	if (&exists( $page ))
	{
		open( PAGE, "< " . $pageFile ) || die( "could not open page file: $pageFile ($!)" ) ;
		while (my $line = <PAGE>)
		{
			print $line ;
		}
		close( PAGE ) ;
	}
	print <<EOF ;
</textarea>
<br>
<input type="submit">
<input type="reset">
</form>
</body>
</html>
EOF
}

sub updatePage
{
        my $page = shift @_ ;
        my $pageFile = &pageFile( $page ) ;
	my $contentType = $ENV{'CONTENT_TYPE'} ;
	defined( $contentType ) || die( "undefined content type" ) ;
	($contentType =~ /^application\/x-www-form-urlencoded$/i) || die( "bad content type: $contentType " ) ;
	my $in ;
	my $contentLength = $ENV{'CONTENT_LENGTH'} ;
	if ($contentLength)
	{
		read( STDIN, $in, $contentLength ) ;
	}
	else
	{
		my $linesep = $/ ;
		$/ = undef ;
		$in = <STDIN> ;
		$/ = $linesep ;
	}
	my %form = parse( $in ) ;
	my $text = $form{"text"} ;
	defined( $text ) || die( "text not defined in form: $in" ) ;
	$text = &normaliseLines( $text ) ;
	open( PAGE, "> " . $pageFile )
		|| die( "could not open page file: $pageFile ($!)" ) ;
	print PAGE $text || die( "could not write to page file: $pageFile ($!)" ) ;
	close( PAGE ) ;
	print "Content-type: text/html\n" ;
	print "\n" ;
	my $viewUrl = viewUrl( $page ) ;
	my $editUrl = editUrl( $page ) ;
	print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>Edited $page</title>
</head>
<body>
<h1>$page</h1>
<p>You have edited <a href="$viewUrl">$page</a>; you should view it to
check that your changes are okay. Alternatively, you may <a
href="$editUrl">edit</a> it again if you like.</p>
</body>
</html>
EOF
}

### CHANGES

sub handleChanges
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = &parse( $queryString ) ;
	my $days = $query{"days"} ;
	my $num = $query{"num"} ;
	if (defined( $days ) || defined( $num ))
	{
		&changesQuery( $days, $num ) ;
	}
	else
	{
		&changesForm() ;
	}
}

sub changesForm
{
	my $frontUrl = &viewUrl( FRONT_PAGE ) ;
	print "Content-type: text/html\n" ;
# todo: put content-type into a constant
	print "\n" ;
	print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>Recent Changes</title>
</head>
<body>
<h1>Recent Changes</h1>
<hr>
<p>This form can be used to query the wiki for a list of pages which have changed recently.</p>
<form method="GET" action="@{[WIKI_CHANGES_ACTION]}">
<table>
<tr><th>Days since the change</th><td><input type="text" name="days"></td></tr>
<tr><th>Number of changes since the change</th><td><input type="text" name="num"></td></tr>
</table>
<input type="submit">
<input type="reset">
</form>
<p>In both fields, blank or 0 means that the criterion is not applied.</p>
<hr>
<p><a href="$frontUrl">@{[FRONT_PAGE]}</a></p>
</body>
</html>
EOF
}

#todo: use subs for common page header and footer

sub changesQuery
{
	my $days = shift @_ ;
	my $num = shift @_ ;
	print "Content-type: text/html\n" ;
	print "\n" ;
        print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>Recent Changes</title>
</head>
<body>
<h1>Recent Changes</h1>
<hr>
<p>Days: $days Number: $num</p>
EOF
#todo: improve the days/number message at the top
	my @changes = &changes( $days, $num ) ;
	print "<ul>\n" ;
	foreach my $page (@changes)
	{
		my $url = viewUrl( $page ) ;
		my $mtime = pageMtime( $page ) ;
		my $mtimestr = gmtime( $mtime ) ;
		print "<li><a href=\"$url\">$page</a> ($mtimestr GMT)</li>\n" ;
	}
	print "</ul>\n" ;
	my $frontUrl = &viewUrl( FRONT_PAGE ) ;
	print <<EOF ;
<hr>
<p><a href="$frontUrl">@{[FRONT_PAGE]}</a></p>
</body>
</html>
EOF
}

sub changes # returns a list of page names, sorted by date
{
	my $days = shift @_ ;
	my $num = shift @_ ;
	# could use opendir, could use glob, could use backticks
	opendir( DIR, WIKI_ROOT ) ;
	my @files = readdir( DIR ) ;
	my @changes ;
	foreach my $file (@files)
	{
#todo: fix
		if ($file =~ /^(.*)\.twici$/)
		{
			my $page = $1 ;
			push( @changes, $page ) ;
		}
	}
	if ( $days )
	{
		my $deadline = time() - ($days * SECS_PER_DAY) ;
		@changes = grep {pageMtime( $_ ) ge $deadline} @changes ;
	}
	@changes = sort { pageMtime( $b ) <=> pageMtime( $a ) } @changes ;
	if ( $num )
	{
		my $len = scalar @changes ;
		my $lim = (($num <= $len) ? $num : $len) - 1 ;
		@changes = @changes[0..$lim] ;
	}
	return @changes ;
}

### BACKLINKS

sub handleBacklinks
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = &parse( $queryString ) ;
	my $page = &cleanPage( $query{"page"} ) ;
	my @backlinks = &backlinks( $page ) ;
	my $pageUrl = &viewUrl( $page ) ;
	print "Content-type: text/html\n" ;
	print "\n" ;
	print DOCTYPE . "\n" ;
	print <<EOF ;
<html>
<head>
<title>Backlinks to $page</title>
</head>
<body>
<h1>Backlinks to <a href="$pageUrl">$page</a></h1>
<hr>
<ul>
EOF
	foreach my $backlink (@backlinks)
	{
		my $backlinkUrl = &viewUrl( $backlink ) ;
		print "<li><a href=\"$backlinkUrl\">$backlink</a></li>\n" ;
	}
	print <<EOF ;
</ul>
<hr>
<p>
EOF
	foreach my $keyPage (KEY_PAGES)
	{
		my $keyUrl = &viewUrl( $keyPage ) ;
		print "| <a href=\"$keyUrl\">$keyPage</a>\n" ;
	}
	print <<EOF ;
|
</p>
</body>
</html>
EOF
}

sub backlinks
{
	my $page = shift @_ ;
        opendir( DIR, WIKI_ROOT ) ;
        my @files = readdir( DIR ) ;
	my @backlinkFiles = grep { &isBacklink( $_, $page ) } @files ;
	my @backlinks = map { substr( $_, 0, -6 ) } @backlinkFiles ;
	return @backlinks ;
}

sub isBacklink
{
	my $file = shift @_ ;
	my $page = shift @_ ;
	if ($file =~ /^(.*)\.twici$/) # factor out a getAllWikiPageFiles sub - use in changes too
	{
		open( PAGE, (WIKI_ROOT . $file) ) ;
		my $hit = 0 ;
		while (!$hit && (my $line = <PAGE>))
		{
			$hit = $line =~ $page ;
		}
		close( PAGE ) ;
		return $hit ;
	}
	else
	{
		return 0 ;
	}
}

### SOURCE

sub handleSource
{
	($ENV{"REQUEST_METHOD"} eq "GET") || die( "script must be accessed by GET" ) ;
	my $queryString = $ENV{"QUERY_STRING"} ;
	my %query = &parse( $queryString ) ;
	my $page = &cleanPage( $query{"page"} ) ;
	my $pageFile = &pageFile( $page ) ;
	open( PAGE, "< " . $pageFile ) || die( "could not open page file" ) ;
	print "Content-type: text/plain\n" ;
	print "\n" ;
	while (my $line = <PAGE>)
	{
		print $line ;
	}
}

### COMMON

sub exists
{
	my $name = shift @_ ;
	my $file = pageFile( $name ) ;
	return -e $file ;
}

sub pageFile
{
	my $page = shift @_ ;
	return &WIKI_ROOT . $page . &WIKI_EXT ;
}

sub viewUrl
{
	my $page = shift @_ ;
	return &WIKI_VIEW_BASE . $page ;
}

sub editUrl
{
	my $page = shift @_ ;
	return &WIKI_EDIT_BASE . $page ;
}

sub sourceUrl
{
	my $page = shift @_ ;
	return &WIKI_SOURCE_BASE . $page ;
}

sub backlinksUrl
{
        my $page = shift @_ ;
        return &WIKI_BACKLINKS_BASE . $page ;
}

sub validateUrl
{
	my $page = shift @_ ;
	my $viewUrl = &viewUrl( $page ) ;
	return "http://validator.w3.org/check?uri=" . $viewUrl ;
}

sub cleanPage
{
	my $page = shift @_ ;
#	$page =~ /&BARE_WIKI_NAME_RE/ || die( "bad page name: $page" ) ; # todo: fix this
	$page =~ /^([A-Z][0-9a-z]*[A-Z][0-9a-zA-Z]*)$/ || die( "bad page name: $page" ) ;
	$page = $1 ;
	return $page ;
}

sub normaliseLines
{
	my $text = shift @_ ;
	return join( "\n", split( /\x0D\x0A|\x0A|\x0D/, $text ) ) . "\n" ;
#	$text =~ s/\x0D\x0A|\x0A|\x0D/\n/ ;
		# todo: make this work
	return $text ;
}

# EMBEDDED_WIKI_NAME_RE => "^([^0-9a-zA-z]*)&WIKI_NAME_RE([^0-9a-zA-z]*)\$"
# WIKI_NAME_RE => "([A-Z][0-9a-z]*[A-Z][0-9a-zA-Z]*)"

sub extractWikiName
{
	my $word = shift @_ ;
	if ($word =~ /^([^0-9a-zA-z]*)([A-Z][0-9a-z]*[A-Z][0-9a-zA-Z]*)([^0-9a-zA-z]*)$/)
	{
		if (wantarray)
		{
			return ($1, $2, $3) ;
		}
		else
		{
			return $2 ;
		}
	}
	else
	{
		return () ;
	}
}

# use constant URL_RE => "([a-zA-Z][0-9a-zA-Z+\-\.]*:.+)" ;
# use constant EMBEDDED_URL_RE => "^([^0-9a-zA-z]*)&URL_RE([^0-9a-zA-z]*)\$" ;

sub extractUrl
{
	my $word = shift @_ ;
	if ($word =~ /^([^0-9a-zA-z]*)([a-zA-Z][0-9a-zA-Z+\-\.]*:.+)([^0-9a-zA-z]*)$/)
	{
		if (wantarray)
		{
			return ($1, $2, $3) ;
		}
		else
		{
			return $2 ;
		}
	}
	else
	{
		return () ;
	}
}

sub parse # todo: replace with CGI.pm call
{
	my $queryString = shift @_ ;
	my @pairs = split( "&", $queryString ) ;
	my %query ;
	foreach my $pair (@pairs)
	{
	        (my $name, my $value) = split( "=", $pair ) ;
		defined( $value ) || die( "bad pair: $pair" ) ;
	        $value =~ tr/+/ / ;
	        $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg ;
	        $value =~ s/\0//g ;
	        $name  =~ s/\0//g ;
	        $query{$name} = $value ;
	}
	return %query ;
}

sub mtime
{
	my $file = shift @_ ;
	my @stat = stat( $file ) ;
	return $stat[9] ;
}

sub pageMtime
{
	my $page = shift @_ ;
	my $file = &WIKI_ROOT . $page . &WIKI_EXT ;
	return &mtime( $file ) ;
}
